@@ -1,10 +1,13 @@
-"""Simple packet creation and parsing."""
-
-from . import triggerlist
+"""
+Simple packet creation and parsing logic.
+"""
 
 import logging
 import random
 import struct
+from collections import OrderedDict
+
+from pypacker import triggerlist
 
 logging.basicConfig(format="%(levelname)s (%(funcName)s): %(message)s")
 #logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
@@ -20,15 +23,15 @@ calcsize = struct.calcsize
 randint = random.randint
 
 
-class Error(Exception):
+class UnpackError(Exception):
 	pass
 
 
-class UnpackError(Error):
+class DissectError(Exception):
 	pass
 
 
-class NeedData(UnpackError):
+class NeedData(Exception):
 	pass
 
 
@@ -38,17 +41,20 @@ class MetaPacket(type):
 	This is done by reading name / format / default out of __hdr__ in every subclass.
 	This configuration is set one time when loading the module (not at instatiation).
 	Actual values are retrieved using "obj.field" notation.
-	CAUTION: list et al are _SHARED_ among all classes! A copy is needed on changes to them.
+
+	CAUTION:
+	- list et al are _SHARED_ among all classes! A copy is needed on changes to them.
 	General note: __new__ is called before __init__
+	- new protocols: don't use header fields having same names as methods in Packet class
 	"""
 
 	def __new__(cls, clsname, clsbases, clsdict):
 		t = type.__new__(cls, clsname, clsbases, clsdict)
-		# all (static+dynamic) header field names (shared)
-		t._hdr_fields = []
-		# all (static+dynamic) as set for performance reasons (shared)
-		t._hdr_fields_set = set()
-		# dictionary of dynamic headers "name" -> TriggerListClass
+		# all (static+dynamic) header field descriptions: name -> format, order needs to be preseved
+		t._hdr_fields = OrderedDict()
+		# all active static+dynamic fields in order
+		t._hdr_fields_active = []
+		# dictionary of dynamic headers: name -> TriggerListClass
 		t._hdr_fields_dyn_dict = {}
 		# get header-infos from subclass: ("name", "format", value)
 		hdrs = getattr(t, "__hdr__", None)
@@ -56,50 +62,62 @@ class MetaPacket(type):
 		if hdrs is not None:
 			#logger.debug("loading meta for: %s, st: %s" % (clsname, st))
 			# all header formats including byte order
-			t._hdr_fmt = [ getattr(t, "__byte_order__", ">")]
+			t._hdr_fmt_order = getattr(t, "__byte_order__", ">")
+			hdr_fmt = [ t._hdr_fmt_order ]
 
 			for hdr in hdrs:
-				#logger.debug("meta: %s -> %s" % (x[0], x[2]))
-				# check if static field (int, bytes etc.) or TriggerList
-				t._hdr_fields.append(hdr[0])
-				t._hdr_fields_set.add(hdr[0])
+				#logger.debug("meta: %s -> %s" % (x[0]))
+				t._hdr_fields[hdr[0]] = hdr[1]
+
+				if hdr[1] is not None:
+				# simple type
+					if hdr[2] is not None:
+					# is active
+						t._hdr_fields_active.append(hdr[0])
+						hdr_fmt.append(hdr[1])
+				else:
+				# assume TriggerList (allways active)
+					t._hdr_fields_active.append(hdr[0])
+					hdr_fmt.append("0s")
 
 				if type(hdr[2]) is not type:
 					# got a simple type, set initial value
 					setattr(t, hdr[0], hdr[2])
-					t._hdr_fmt.append(hdr[1])
 				else:
-					# got a TriggerList
-					#logger.debug("got dynamic field: %s=%s" % (hdr[0], hdr[2]))
-					t._hdr_fmt.append(None)
+					# assume TriggerList
 					# remmember for lazy instantiation
 					t._hdr_fields_dyn_dict[hdr[0]] = hdr[2]
 
 			# current format bytestring as string for convenience
-			t._hdr_fmtstr = "".join(v for v in t._hdr_fmt if v is not None)
-			#logger.debug("formatstring is: %s" % t._hdr_fmtstr)
-			t._hdr_len = calcsize(t._hdr_fmtstr)
-			# body as raw byte string (None if handler present)
-			t._data = b""
+			t._hdr_fmt = struct.Struct("".join(v for v in hdr_fmt))
+			#logger.debug("formatstring is: %s" % hdr_fmt)
+			# body as raw byte string (None if handler is present)
+			t._body_bytes = b""
 			# name of the attribute which holds the object representing the body aka the body handler
 			t._bodytypename = None
+			# indicates something went wrong on this layer parsing NEXT upper layer
+			t._parsefail = False
 			# callback to the next lower layer (eg for checksum on IP->TCP/UDP)
 			t._callback = None
 			# track changes to header values and data: This is needed for layers like TCP for
 			# checksum-recalculation. Set to "True" on changes to header/body values, set to False on "bin()"
 			## track changes to header values
 			t._header_changed = False
-			## track changes to header format. This will happen when adding new header or using TriggerLists
+			## track changes to header format. This will happen wg when changing TriggerLists
 			t._header_format_changed = False
 			## track changes to body value like [None | bytes | body-handler] -> [None | bytes | body-handler]
 			t._body_changed = False
 			# cache header for performance reasons, this will be set to None on every change to header values
 			t._header_cached = None
-			# objects which get notified on changes on _header_ values via "__setattr__()" (shared)
+			# objects which get notified on changes on header or body (shared)
 			# TODO: use sets here
 			t._changelistener = []
 			# lazy handler data, format: [name, class, bytes]
 			t._lazy_handler_data = None
+			# indicates the most top layer until which should be unpacked (vs. lazy parsing = just 1st layer)
+			t._target_unpack_clz = None
+			# indicates if active-field list for tracking header infos is still shared
+			t._hdrlist_original = True
 		return t
 
 
@@ -140,13 +158,15 @@ class Packet(object, metaclass=MetaPacket):
 	============
 
 		- Auto-decoding of headers via given format-patterns (defined via __hdr__)
-		- Auto-decoding of body-handlers like IP -> parse IP-data -> add TCP-handler to IP -> parse TCP-data..
+		- Auto-decoding of body-handlers (like IP -> parse IP-data -> add TCP-handler to IP -> parse TCP-data..)
 		- Access of fields via "layer1.key" notation
-		- Some members can be set/retrieved using convenient string-represenations beneath the
-			byte-represenation (see Ethernet or IP). This is done by appending "_s" to the attributename:
-			obj.key_s = "stringval"
-			string_var = obj.key_s
-			Convenient access should be set via: varname_s = pypacker.Packet._get_property_XXX("varname")
+		- Convenient access for standard types (MAC, IP address) using string-represenations
+			This is done by appending "_s" to the attributename:
+			ip.src_s = "127.0.0.1"
+			ip_src_bytes = ip.src
+
+			Implementation info:
+			Convenient access should be set via varname_s = pypacker.Packet._get_property_XXX("varname")
 		- Access of higher layers via layer1.layer2.layerX or "layer1[layerX]" notation
 		- Concatination via "layer1 + layer2 + layerX"
 		- There are two types of headers:
@@ -162,9 +182,10 @@ class Packet(object, metaclass=MetaPacket):
 				can be overwritten (see ip.py)
 		- Header-values with length < 1 Byte should be set by using properties
 		- Header formats can not be updated directly
+		- Static fields can be deactivated via setting None "obj.field = None", re-activate by setting a value
 		- Ability to check direction to other Packets via "direction()"
-		- Generic callback for rare cases eg where upper layer needs
-			to know about lower ones (like TCP->IP for checksum calculation)
+		- Generic callback to lower layers for rare cases eg where upper layer needs
+			to know about lower ones (like TCP (higher layer) ->IP (lower layer) for checksum calculation)
 		- No correction of given raw packet-data eg checksums when creating a packet from it
 			(exception: if the packet can't be build without correct data -> raise exception).
 			The internal state will only be updated on changes to headers or data.
@@ -198,7 +219,7 @@ class Packet(object, metaclass=MetaPacket):
 	>>> foo
 	Foo(baz="whee", foo=7, bar=3)
 	>>> Foo(b"hello, world!")
-	Foo(baz=" wor", foo=1751477356L, bar=28460, data="ld!")
+	Foo(baz=" wor", foo=1751477356L, bar=28460, body_bytes="ld!")
 	"""
 
 	"""Dict for saving body datahandler globaly: { Classname : {id : HandlerClass} }"""
@@ -206,11 +227,8 @@ class Packet(object, metaclass=MetaPacket):
 	"""Basic types allowed for header-values"""
 	__TYPES_ALLOWED_BASIC = set([bytes, int, float])
 	"""Constants for Packet-directons"""
-	"""Same direction"""
 	DIR_SAME	= 1
-	"""Reversed direction"""
 	DIR_REV		= 2
-	"""Direction unknown"""
 	DIR_UNKNOWN	= 3
 
 	def __init__(self, *args, **kwargs):
@@ -227,12 +245,20 @@ class Packet(object, metaclass=MetaPacket):
 			# Allowing default-values giving empty buffer would lead to confusion:
 			# there is no way do disambiguate "no body" from "default value set".
 			# Empty buffer in A for subhandler B (this Packet) should lead to A
-			# having (data=b"", bodyhandler=None)
+			# having (body_bytes=b"", bodyhandler=None)
 			if len(args[0]) == 0:
 				raise NeedData("Empty buffer given!")
+			elif len(args) > 1:
+				# additional parameters are only given by packet-class itself
+				self._target_unpack_clz = args[1]._target_unpack_clz
 
 			# this is called on the extended class if present
-			self._dissect(args[0])
+			try:
+				self._dissect(args[0])
+			except Exception as e:
+				self._parsefail = True
+				logger.exception("could not dissect, trying to unpack after all..")
+				# go on, we can still try to unpack...
 
 			try:
 				# assign values to this packet
@@ -240,7 +266,6 @@ class Packet(object, metaclass=MetaPacket):
 			except UnpackError as ex:
 				raise UnpackError("could not unpack %s: %s" % (self.__class__.__name__, ex))
 		else:
-			# n headerfields given to set (n >= 0)
 			# additional parameters given, those overwrite the class-based attributes
 			#logger.debug("New Packet with keyword args (%s)" % self.__class__.__name__)
 			for k, v in kwargs.items():
@@ -259,9 +284,9 @@ class Packet(object, metaclass=MetaPacket):
 	def __len__(self):
 		"""Return total length (= header + all upper layer data) in bytes."""
 
-		if self._data is not None:
+		if self._body_bytes is not None:
 			#logger.debug("returning length from raw bytes in %s" % self.__class__.__name__)
-			return self.hdr_len + len(self._data)
+			return self.hdr_len + len(self._body_bytes)
 		else:
 			try:
 				# avoid unneeded parsing
@@ -273,30 +298,25 @@ class Packet(object, metaclass=MetaPacket):
 				return self.hdr_len + len(self.__getattribute__(self._bodytypename))
 
 	#
-	# Public access to header length: keep it uptodate
+	# public access to header length: keep it uptodate
 	#
 	def __get_hdrlen(self):
 		# format changed: recalculate length
 		if self._header_format_changed:
-			self._update_fmtstr()
-		return self._hdr_len
+			self._update_fmt()
+		return self._hdr_fmt.size
 
+	# update format if needed and return actual header size
 	hdr_len = property(__get_hdrlen)
+	header_len = property(__get_hdrlen)
 
-	#
-	# Public access to format string: keep it uptodate
-	#
-	def __get_fmtstr(self):
-		# format changed: rebuild format string
-		if self._header_format_changed:
-			self._update_fmtstr()
-		return self._hdr_fmtstr
-
-	hdr_fmtstr = property(__get_fmtstr)
+	# non-intrusive version of "hdr_len": doesn't change anything, format needs to be uptodate to return correct length
+	# TODO: remove if unneeded
+	_hdr_len = property(lambda v: v._hdr_fmt.size)
 
 	# Two types of data can be set: raw bytes or handler, use property for convenient access
-	# The following assumption must be fullfilled: (handler=obj, data=None) OR (handler=None, data=b"")
-	def __get_data(self):
+	# The following assumption must be fullfilled: (handler=obj, body_bytes=None) OR (handler=None, body_bytes=b"")
+	def __get_bodybytes(self):
 		"""
 		Return raw data bytes or handler bytes (including all upper layers) if present.
 		This is the same as calling bin() but excluding this header and without resetting changed-status.
@@ -307,13 +327,13 @@ class Packet(object, metaclass=MetaPacket):
 		elif self._bodytypename is not None:
 		# some handler was set
 			hndl = object.__getattribute__(self, self._bodytypename)
-			return hndl.pack_hdr() + hndl.data
+			return hndl._pack_header() + hndl.__get_bodybytes()
 		# return raw bytes
 		else:
-			return self._data
+			return self._body_bytes
 
-	def __set_data(self, value):
-		"""Allow obj.data = [None | b"" | Packet]. None will reset any body handler."""
+	def __set_body(self, value):
+		"""Allow obj._body_bytes = [None | b"" | Packet]. None will reset any body handler."""
 		if type(value) is bytes:
 			if self._bodytypename is not None:
 			# reset all handler data
@@ -321,58 +341,46 @@ class Packet(object, metaclass=MetaPacket):
 			# track changes to raw data
 			self._body_changed = True
 			#logger.debug("setting new raw data: %s" % value)
-			self._data = value
-			self._handle_mod("data", value)
+			self._body_bytes = value
+			self._handle_mod("body_bytes", value)
 		else:
 			# set body handler (can be None), assume value is a Packet
 			# this will set body changed status to True
 			self._set_bodyhandler(value)
+		self.__notity_changelistener()
 
-	data = property(__get_data, __set_data)
-
-	def __get_body_hndl(self, searching_for=None):
-		"""Return handler object or None if not present."""
-		if self._lazy_handler_data is not None:
-		# parse lazy handler data on the next layer
-			# this is more likely
-			if searching_for is None:
-				return self.__getattr__(self._lazy_handler_data[0])
-			else:
-				# TODO: avoid re-init of lazy data if we know that handler in question is not present here
-				pass
-		elif self._bodytypename is not None:
-		# body handler allready parsed
-			return self.__getattribute__(self._bodytypename)
-		else:
-		# nope, chuck testa
-			#logger.debug("returning None")
-			return None
-
-	_body_handler = property(__get_body_hndl)
+	# return body data as raw bytes (deprecated)
+	data = property(__get_bodybytes, __set_body)
+	# get (bytes) and set (bytes, handler) for body
+	body_bytes = property(__get_bodybytes, __set_body)
 
 	def __getattr__(self, k):
 		"""
 		Gets called if there are no fields matching the name k. Check if we got
-		lazy handler data set which must get parsed.
+		lazy handler data or lazy dynamic fields set which must now me initiated.
 		"""
 		try:
 			if self._lazy_handler_data[0] == k:
 			# lazy handler data was set, parse lazy handler data now!
 				#logger.debug("lazy parsing handler: %s" % k)
 				handler_data = self._lazy_handler_data
-				#buf = handler_data[2][handler_data[3] : handler_data[4]]
+				# TOOD: this is a bit redundant: see _parse_handler()
 
 				try:
-				# instantiate handler class using buffer
-					type_instance = handler_data[1]( handler_data[2] )
+				# instantiate handler class using lazy data buffer
+				# See _parse_handler() for 2nd place where handler instnation takes place
+					type_instance = handler_data[1](handler_data[2], self)
 					self._set_bodyhandler( type_instance )
 				except Exception as e:
-					logger.warning("could not lazy-parse handler %s (len: %d): %s" %
-						(handler_data[1], len(handler_data[2]), e))
-					# TODO: set via "self.data"?
-					self._bodytypename = None
-					self._data = handler_data[2]
 					type_instance = None
+					logger.exception("could not lazy-parse handler %s (len: %d): %s, there could be 2 reasons for this:" +
+						"1) packet was malformed 2) parsing-code is buggy" % (handler_data[1], len(handler_data[2]), e))
+					self._bodytypename = None
+					self._body_bytes = handler_data[2]
+					self._parsefail = True
+
+					# TODO: remove this to ignore parse errors (set raw bytes after all)
+					#raise Exception("%r" % e)
 
 				self._lazy_handler_data = None
 				# this was a lazy init: same as direct parsing -> no body change
@@ -380,8 +388,8 @@ class Packet(object, metaclass=MetaPacket):
 
 				return type_instance
 		except TypeError:
+			# no lazy handler data, ignore
 			pass
-		# lazy data present but key not matching
 
 		# init dynamic fields
 		if k in self._hdr_fields_dyn_dict:
@@ -391,27 +399,53 @@ class Packet(object, metaclass=MetaPacket):
 
 		#logger.warning("unable to find: %s" % k)
 		# nope..not found..
-		raise AttributeError
+		raise AttributeError("Can't find Attribute in %r: %s" % (self.__class__, k))
+
+	# indicates something went wrong on this layer parsing NEXT upper layer
+	parsefail = property(lambda _: _._parsefail)
+
+	def _deactivate_hdr(self, hdr):
+		# deactivating is less costly than activating
+		#logger.debug("de-activating: %s" % hdr)
+		if self._hdrlist_original:
+			self._hdr_fields_active = list(self._hdr_fields_active)
+			self._hdrlist_original = False
+		self._hdr_fields_active.remove(hdr)
+		self._header_format_changed = True
+
+	def _activate_hdr(self, hdr):
+		#logger.debug("activating: %s" % hdr)
+		# we need the correct order: use _hdr_fields
+		# assuming field was already set to "None"
+		self._hdr_fields_active = [name for name in self._hdr_fields if name in self._hdr_fields_active + [hdr]]
+		self._hdrlist_original = False
+		self._header_format_changed = True
 
 	def __setattr__(self, k, v):
 		"""
 		Set value of an attribute "k" via "a.k=v".
 		"""
-		# Track changes to header fields
-		if k in self._hdr_fields_set:
-			#logger.debug("setting attribute: %s: %s->%s" % (self.__class__, k, v))
+		#logger.debug("setting attribute: %s: %s->%s" % (self.__class__, k, v))
+		# track changes to header fields
+		if k in self._hdr_fields:
+			#logger.debug("setting field attribute: %s: %s->%s" % (self.__class__, k, v))
 			self._header_changed = True
 			self._header_cached = None
 
-			# set allowed on basic types
 			if not k in self._hdr_fields_dyn_dict:
-				#logger.debug("setting static header value: %s=%s" % (k,v))
 				object.__setattr__(self, k, v)
+
+				# check for activated/deactivated header
+				# TODO: use dicts?
+				if v is None and k in self._hdr_fields_active:
+					self._deactivate_hdr(k)
+				elif v is not None and not k in self._hdr_fields_active:
+					self._activate_hdr(k)
 			else:
 			# TriggerList: avoid overwriting dynamic fields when using keyword constructor Class(key=val)
 			# triggerlistObj = [ b"" | (KEY_X, VAL) | [(KEY_X, VAL), ...]] => clear current
 			# list and add value.
-				#logger.debug("adding triggerlist values: %s=%s" % (k,v))
+				#logger.debug("got obj.triggerlist = b'': adding triggerlist values: %s=%s" % (k,v))
 				# this will trigger a lazy init
 				header_val = getattr(self, k)
 				del header_val[:]
@@ -430,6 +464,7 @@ class Packet(object, metaclass=MetaPacket):
 		"""
 		Handle modification of packet fields like "packet.field = value". This enables
 		updating eg length fields when setting data values. Default imeplementation does nothing.
+		Needed for: Changes to packets which affect other fields eg length fields (see ip.IP).
 		NOTE: in order to react on TriggerList changes create a specialized one and
 		overwrite _handle_mod() there.
 		"""
@@ -442,13 +477,13 @@ class Packet(object, metaclass=MetaPacket):
 
 		k -- Packet-type to search for like Ethernet, IP, TCP etc.
 		"""
-		# TODO: testcase: access to non existend handler -> retrieve data (eth.http -> eth.data)
 		p_instance = self
+		# set most top layer to be unpacked, __getattr__() could be called unpacking lazy data
+		self._target_unpack_clz = k
 
 		while not type(p_instance) is k:
-			# this will auto-parse lazy handler data
-			# __get_body_hndl(skip_lazy_init=True)
-			p_instance = p_instance._body_handler
+			# this will auto-parse lazy handler data via _get_bodyhandler()
+			p_instance = p_instance._get_bodyhandler()
 
 			if p_instance is None:
 				break
@@ -456,6 +491,35 @@ class Packet(object, metaclass=MetaPacket):
 		#logger.debug("returning found packet-handler: %s->%s" % (type(self), type(p_instance)))
 		return p_instance
 
+	def __iter__(self):
+		"""
+		Iterate over every layer starting with this ending at last/highest one
+		"""
+		p_instance = self
+		# assume string class never gets found
+		self._target_unpack_clz = str.__class__
+
+		while not p_instance is None:
+			yield p_instance
+			# this will auto-parse lazy handler data via _get_bodyhandler()
+			p_instance = p_instance._get_bodyhandler()
+
+			if p_instance is None:
+				break
+
+	def dissect_full(self):
+		"""
+		Recursive unpack ALL data inlcuding lazy header etc up to highest layer inlcuding danymic fields.
+		"""
+		for hdr in self._hdr_fields_dyn_dict:
+			self.__getattribute__(hdr)._lazy_dissect()
+
+		try:
+			self._get_bodyhandler().dissect_full()
+		except AttributeError:
+			# no handler present
+			pass
+
 	def __add__(self, packet_to_add):
 		"""
 		Handle concatination of layers like "Ethernet + IP + TCP" and make them accessible
@@ -470,16 +534,18 @@ class Packet(object, metaclass=MetaPacket):
 
 		# get highest layer from this packet
 		highest_layer = self
+		# unpack all layer, assuming string class will be never found
+		self._target_unpack_clz = str.__class__
 
 		while highest_layer is not None:
 			if highest_layer._bodytypename is not None:
-				highest_layer = highest_layer._body_handler
+				highest_layer = highest_layer._get_bodyhandler()
 			else:
 				break
 
 		# connect callback from lower (this packet, highest_layer) to upper (packet_to_add) layer eg IP->TCP
 		packet_to_add._callback = highest_layer._callback_impl
-		highest_layer.data = packet_to_add
+		highest_layer._set_bodyhandler(packet_to_add)
 
 		return self
 
@@ -493,8 +559,9 @@ class Packet(object, metaclass=MetaPacket):
 		# create key=value descriptions
 		# this will lazy init dynamic fields
 		l = [ "%s=%r" % (k, getattr(self, k)) for k in self._hdr_fields]
-		if self._data is not None:
-			l.append("data=%r" % self._data)
+		# no handler present
+		if self._body_bytes is not None:
+			l.append("bytes=%r" % self.body_bytes)
 		else:
 			# assume bodyhandler is set
 			#l.append("handler=%s" % self.__getattribute__(self._bodytypename).__class__)
@@ -514,9 +581,6 @@ class Packet(object, metaclass=MetaPacket):
 		"""Create a get/set-property for an IP4 address as string-representation."""
 		return property(lambda self: ip4_bytes_to_str(object.__getattribute__(self, var)),
 		lambda self, val: self.__setattr__(var, ip4_str_to_bytes(val)))
-	#
-	#
-	#
 
 	def _unpack(self, buf):
 		"""
@@ -528,26 +592,32 @@ class Packet(object, metaclass=MetaPacket):
 
 		buf -- the buffer to be parsed
 		"""
-		cnt = 1
-
 		try:
 			# calling "self.hdr_len" will update format for dynamic fields
 			self._header_cached = buf[:self.hdr_len]
+			hdr_unpacked = self._hdr_fmt.unpack(self._header_cached)
 
-			for k, v in zip(self._hdr_fields,
-					unpack(self.hdr_fmtstr, self._header_cached)):
-				# only set non-TriggerList fields
-				if self._hdr_fmt[cnt] is not None:
-					object.__setattr__(self, k, v)
+			#for name_bytes in zip(self._hdr_fields_active, hdr_unpacked):
+			#	if self._hdr_fields[name_bytes[0]] is not None:
+			##		#logger.debug("setting value: %s -> %s" % (name_bytes[0], name_bytes[1]))
+			#		object.__setattr__(self, name_bytes[0], name_bytes[1])
+			cnt = 0
+			for name in self._hdr_fields_active:
+				if self._hdr_fields[name] is not None:
+			#		#logger.debug("setting value: %s -> %s" % (name_bytes[0], name_bytes[1]))
+					object.__setattr__(self, name, hdr_unpacked[cnt])
 				cnt += 1
-		except IndexError:
-			raise NeedData("Not enough data to unpack: buf %d < %d" % (len(buf), self.hdr_len))
+		except Exception:
+			raise UnpackError("could not unpack, format/hdr/active: %s/%r/%r/%s" % (self._hdr_fmt.format,
+												self._hdr_fields,
+												self._hdr_fields_active,
+												self._header_cached))
 
 		# extending class didn't set data itself, set raw data
 		if not self._body_changed:
-			self._data = buf[self._hdr_len:]
+			self._body_bytes = buf[self._hdr_fmt.size:]
 
-		#logger.debug("header: %s, body: %s" % (self._hdr_fmtstr, self.data))
+		#logger.debug("header: %s, body: %s" % (self._hdr_fmt, self.body_bytes))
 		# reset the changed-flags: original unpacked value = no changes
 		self.__reset_changed()
 
@@ -588,11 +658,11 @@ class Packet(object, metaclass=MetaPacket):
 				new_packet = new_layer
 
 			# next layer to be copied
-			if current_hndl._body_handler is None:
+			if current_hndl._get_bodyhandler() is None:
 				# upper layer reached: set raw bytes
-				new_layer.data = current_hndl.data
+				new_layer._set_body_bytes( current_hndl._get_bodybytes() )
 
-			current_hndl = current_hndl._body_handler
+			current_hndl = current_hndl._get_bodyhandler()
 
 		return new_packet
 
@@ -609,9 +679,9 @@ class Packet(object, metaclass=MetaPacket):
 	def _parse_handler(self, hndl_type, buffer):
 		"""
 		Called by overwritten "_dissect()":
-		Parse the handler using the given buffer and set it using _set_bodyhandler() later on (Lazy
+		Initiate the handler-parser using the given buffer and set it using _set_bodyhandler() later on (lazy
 		init). This will use the calling class and given handler type to retrieve the resulting handler.
-		On any error this will set raw bytes given for data.
+		On any error this will set raw bytes given for body data.
 
 		hndl_type -- A value to place the handler in the handler-dict like
 			dict[Class.__name__][hndl_type] (eg type-id, port-number)
@@ -623,21 +693,35 @@ class Packet(object, metaclass=MetaPacket):
 			return
 
 		try:
+			if self._target_unpack_clz is None or self._target_unpack_clz is self.__class__:
 			# set lazy handler data, __getattr__() will be called on access to handler
-			clz = Packet._handler[self.__class__.__name__][hndl_type]
-			clz_name = clz.__name__.lower()
-			#logger.debug("setting handler name: %s -> %s" % (self.__class__.__name__, clz_name))
-			self._lazy_handler_data = [clz_name, clz, buffer]
-			# set name allthough we don't set a handler (needed for direction() et al)
-			self._bodytypename = clz_name
-			# avoid setting data by "_unpack"
-			self._body_changed = True
-			self._data = None
-		except KeyError as e:
-			#logger.debug("can't set lazy handler data type %s in %s: type unknown" %
-			#	(str(hndl_type), self.__class__.__name__))
+				clz = Packet._handler[self.__class__.__name__][hndl_type]
+				clz_name = clz.__name__.lower()
+				#logger.debug("setting handler name: %s -> %s" % (self.__class__.__name__, clz_name))
+				self._lazy_handler_data = [clz_name, clz, buffer]
+				# set name allthough we don't set a handler (needed for direction() et al)
+				self._bodytypename = clz_name
+				# avoid setting data by "_unpack"
+				self._body_changed = True
+				self._body_bytes = None
+			else:
+			# continue parsing layers, happens von "__getitem__()": avoid unneeded lazy-data creation
+			# if specific class must be found
+				#logger.debug("--------> direct unpacking in: %s" % (self.__class__.__name__))
+				type_instance = Packet._handler[self.__class__.__name__][hndl_type](buffer, self)
+				self._set_bodyhandler(type_instance)
+		except KeyError:
+			logger.info("unknown type for %s: %d, feel free to implement" % (self.__class__, hndl_type))
+			self.body_bytes = buffer
+			self._parsefail = True
+		except Exception as e:
+			logger.exception("can't set handler data, type/lazy: %s/%s:" %
+				(str(hndl_type), self._target_unpack_clz is None or self._target_unpack_clz is self.__class__))
 			# set raw bytes as data (eg handler class not found)
-			self.data = buffer
+			self.body_bytes = buffer
+			self._parsefail = True
+			# TODO: remove this to ignore parse errors (set raw bytes after all)
+			#raise Exception("%r" % e)
 
 	def direction(self, packet2):
 		"""
@@ -657,7 +741,7 @@ class Packet(object, metaclass=MetaPacket):
 		try:
 			# check upper layers and combine current result
 			#logger.debug("direction? checking next layer")
-			return dir_ext & self._body_handler.direction( next._body_handler )
+			return dir_ext & self._get_bodyhandler().direction( next._get_bodyhandler() )
 		except AttributeError:
 			# one of both _bodytypename was None
 			# Example: TCP ACK (last step of handshake, no payload) <-> TCP ACK + Telnet
@@ -675,42 +759,50 @@ class Packet(object, metaclass=MetaPacket):
 		"""
 		return self.direction(packet2) & direction == direction
 
-	def _update_fmtstr(self):
+	def _update_fmt(self):
 		"""
 		Update header format string and length using current fields.
-		NOTE: only called if format has changed eg after addin/removing header fields,
-		changes in TriggerList etc.
+		This will also update the active headers.
+		NOTE: only called if format has changed eg on changes in TriggerList etc.
 		"""
 		#logger.debug("updating format")
 		# byte-order is set via first character
-		hdr_fmt_tmp = [ self._hdr_fmt[0] ]
+		hdr_fmt_tmp = [ self._hdr_fmt_order ]
+		#logger.debug("active fields: %r" % self._hdr_fields_active)
 
 		# we need to preserve the order of formats / fields
-		for idx, field in enumerate(self._hdr_fields):
-			# Three options:
-			# - value bytes			-> add given format or calculate by length
-			# - value TriggerList		(found via format None)
-			#	- type Packet		-> a TriggerList of packets, reassemble formats
-			#	- type tuple		-> a TriggerList of tuples, call "reassemble" and use format "s"
-			#logger.debug("format update with field/type/val: %s/%s/%s" % (field, type(val), val))
-			if self._hdr_fmt[1 + idx] is not None:				# bytes/int/float
-				hdr_fmt_tmp.append( self._hdr_fmt[1 + idx] )		# skip byte-order character
+		for name in self._hdr_fields_active:
+			#logger.debug("format update with field/format: %s/%s" % (name, self._hdr_fields[name]))
+			# two options:
+			# - value bytes				-> add given format
+			# - value TriggerList			(found via format None) -> call bin()
+			if self._hdr_fields[name] is not None:				# bytes/int/float
+				hdr_fmt_tmp.append( self._hdr_fields[name] )		# skip byte-order character
 			else:								# assume TriggerList
 				try:
-					val = object.__getattribute__(self, field).bin()
-
-					if len(val) > 0:
-						hdr_fmt_tmp.append( "%ds" % len(val) )
+					val = object.__getattribute__(self, name).bin()
+					hdr_fmt_tmp.append( "%ds" % len(val) )
 				except AttributeError:
+					hdr_fmt_tmp.append("0s")
 					# dynamic field not yet initiated = no value parsed = not needed: skip
 					continue
 
-		hdr_fmt_tmp = "".join(hdr_fmt_tmp)
-
-		# update header info, avoid recursive calls
-		self._hdr_fmtstr = hdr_fmt_tmp
+		#logger.debug("format update class/hdr: %s/%r" % (self.__class__.__name__, self._hdr_fields))
+		self._hdr_fmt = struct.Struct("".join(hdr_fmt_tmp))
 		self._header_format_changed = False
-		self._hdr_len = calcsize(hdr_fmt_tmp)
+
+	def _get_bodyhandler(self):
+		"""Return handler object or None if not present."""
+		if self._lazy_handler_data is not None:
+		# parse lazy handler data on the next layer
+			return self.__getattr__(self._lazy_handler_data[0])
+		elif self._bodytypename is not None:
+		# body handler allready parsed
+			return self.__getattribute__(self._bodytypename)
+		else:
+		# nope, chuck testa
+			#logger.debug("returning None")
+			return None
 
 	def _set_bodyhandler(self, hndl):
 		"""
@@ -726,22 +818,25 @@ class Packet(object, metaclass=MetaPacket):
 			raise Error("can't set handler which is not a Packet")
 
 		if hndl is None:
-		# switch (handler=obj, data=None) to (handler=None, data=b'')
+		# switch (handler=obj, body_bytes=None) to (handler=None, body_bytes=b'')
 			self._bodytypename = None
-			# avoid (data=None, handler=None)
-			self._data = b""
+			# avoid (body_bytes=None, handler=None)
+			self._body_bytes = b""
 		else:
 		# set a new body handler
 			# associate ip, arp etc with handler-instance to call "ether.ip", "ip.tcp" etc
 			self._bodytypename = hndl.__class__.__name__.lower()
 			hndl._callback = self._callback_impl
 			object.__setattr__(self, self._bodytypename, hndl)
-			self._data = None
+			self._body_bytes = None
 
 		self._lazy_handler_data	= None
 		# new body handler means body data changed
 		self._body_changed = True
 
+	# get/set body handler or None. Note: this will force lazy parsing when reading
+	body_handler = property(_get_bodyhandler, _set_bodyhandler)
+
 	def bin(self):
 		"""
 		Return this header and body (including all upper layers) as byte string
@@ -754,53 +849,53 @@ class Packet(object, metaclass=MetaPacket):
 			data_tmp = self._lazy_handler_data[2]
 		elif self._bodytypename is not None:
 			# handler allready parsed
-			data_tmp = self._body_handler.bin()
+			data_tmp = self._get_bodyhandler().bin()
 		else:
 			# raw bytes
-			data_tmp = self._data
+			data_tmp = self._body_bytes
 		# now every layer got informed about our status, reset it
 		self.__reset_changed()
-		return self.pack_hdr() + data_tmp
+		return self._pack_header() + data_tmp
 
-	def pack_hdr(self):
+	def _pack_header(self):
 		"""
 		Return header as byte string.
 		"""
 		if self._header_format_changed:
-			self._update_fmtstr()
+			self._update_fmt()
 		# return cached data if nothing changed
 		elif self._header_cached is not None:
 			#logger.warning("returning cached header (hdr changed=%s): %s->%s" %\
-			#	(self.header_changed, self.__class__.__name__, self._header_cached))
+			#	(self._header_changed, self.__class__.__name__, self._header_cached))
 			return self._header_cached
 
-		try:
-			hdr_bytes = []
+		hdr_bytes = []
 
-			for idx, field in enumerate(self._hdr_fields):
+		for name in self._hdr_fields_active:
+			# two options:
+			# - value bytes			-> add given bytes
+			# - value TriggerList		(found via format None) -> call bin()
+			#logger.debug("packing header with field/format/val: %s/%s/%s" % (name, self._hdr_fields[name], object.__getattribute__(self, name)))
+			if self._hdr_fields[name] is not None:			# bytes/int/float
+				val = object.__getattribute__(self, name)
+				hdr_bytes.append(val)
+			else:							# assume TriggerList
 				try:
-					val = object.__getattribute__(self, field)
+					val = object.__getattribute__(self, name)
 				except AttributeError:
+					hdr_bytes.append(b"")
 					# dynamic field not yet initiated: skip
 					continue
-				# three options:
-				# - value bytes			-> add given bytes
-				# - value TriggerList		(found via format None)
-				#	- type Packet		-> a TriggerList of packets, reassemble bytes
-				#	- type tuple		-> a Triggerlist of tuples, call "pack"
-				#logger.debug("packing header with field/type/val: %s/%s/%s" % (field, type(val), val))
-				if self._hdr_fmt[1 + idx] is not None:		# bytes/int/float
-					hdr_bytes.append( val )
-				else:						# assume TriggerList
-					val = val.bin()
-					if len(val) > 0:
-						hdr_bytes.append( val )
-			#logger.debug("header bytes for %s: %s = %s" % (self.__class__.__name__, self._hdr_fmtstr, hdr_bytes))
-			self._header_cached = pack( self._hdr_fmtstr, *hdr_bytes )
 
-			return self._header_cached
-		except Exception as e:
-			logger.warning("error while packing header: %s" % e)
+				hdr_bytes.append( val.bin() )
+		#logger.debug("header bytes for %s: %s = %s" % (self.__class__.__name__, self._hdr_fmt.format, hdr_bytes))
+		self._header_cached = self._hdr_fmt.pack(*hdr_bytes)
+		#logger.debug("cached header: %s" % self._header_cached)
+
+		return self._header_cached
+
+	# readonly access to header
+	header_bytes = property(_pack_header)
 
 	def _changed(self):
 		"""
@@ -816,7 +911,7 @@ class Packet(object, metaclass=MetaPacket):
 				break
 			elif p_instance._lazy_handler_data is None:
 			# one layer up, stop if next layer is not yet initiated which means: no change
-				p_instance = p_instance._body_handler
+				p_instance = p_instance._get_bodyhandler()
 			else:
 			# nothing changed upwards: lazy handler data still present/nothing got parsed
 				p_instance = None
@@ -861,26 +956,22 @@ class Packet(object, metaclass=MetaPacket):
 			try:
 				listener_cb(self)
 			except Exception as e:
-				logger.warning("error when informing listener: %s" % e)
-				#pass
+				logger.exception("error when informing listener: %r" % e)
 
 	def __load_handler(clz, clz_add, handler):
 		"""
 		Load Packet handler using a shared dictionary.
 
-		clz_add -- class to be added
+		clz_add -- class for which handler has to be added
 		handler -- dict of handlers to be set like { id : class }, id can be a tuple of values
 		"""
 		clz_name = clz_add.__name__
 
-		try:
-			Packet._handler[clz_name]
-			logger.debug(">>> handler already loaded: %s (%d)" % clz_add)
+		if clz_name in Packet._handler:
+			logger.debug("handler already loaded: %r" % clz_name)
 			return
-		except KeyError:
-			pass
 
-		logger.debug("adding classes as handler: [%s] = %s" % (clz_add, handler))
+		logger.debug("adding classes as handler: [%r] = %r" % (clz_add, handler))
 
 		Packet._handler[clz_name] = {}
 
@@ -906,7 +997,7 @@ def mac_str_to_bytes(mac_str):
 
 def mac_bytes_to_str(mac_bytes):
 	"""Convert mac address from byte representation to AA:BB:CC:DD:EE:FF."""
-	return "%02x:%02x:%02x:%02x:%02x:%02x" % unpack("BBBBBB", mac_bytes)
+	return "%02X:%02X:%02X:%02X:%02X:%02X" % unpack("BBBBBB", mac_bytes)
 
 
 def get_rnd_mac():
